#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

layout(location = 0) in vec3 position_world;
layout(location = 1) in vec3 normal_world;
layout(location = 2) in vec2 uv;


out vec4 color;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular
uniform float specularAlpha;
uniform sampler2D diffuseTexture;
uniform sampler2D directionalShadowMap;
uniform mat4 directionalLightTransform;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;

uniform samplerCube EnvMap;

//normalised device coordinated
float CalcDirectionalShadowFactor(DirectionalLight light)
{
	vec4 DirectionalLightSpacePos = directionalLightTransform * vec4(position_world, 1.0);
	vec3 projCoords = DirectionalLightSpacePos.xyz / DirectionalLightSpacePos.w;
	projCoords = (projCoords * 0.5) + 0.5;

	//depth
	float current = projCoords.z;

	//=============
	vec3 normal = normalize(normal_world);
	vec3 lightDir = normalize(light.direction);

	//depthbias
	float bias = max(0.00005 * (1.0 - dot(normal, lightDir)), 0.000005);

	//PCF
	float shadow = 0.0f;
	vec2 texelSize = 1.0f / textureSize(directionalShadowMap, 0);
	for(int x = -1; x <= 1; ++x)
  {
		for(int y = -1; y <= 1; ++y)
		{
			float pcfDepth = texture(directionalShadowMap, projCoords.xy + vec2(x,y) * texelSize).r;
			shadow += current -bias > pcfDepth ? 1.0f : 0.0f;
		}
	}

	//averaging the surrounding pixels
	shadow /= 9.0f;

	//turning off shadows that are further away than the far plane
	if(projCoords.z > 1.0f)
	{
		shadow = 0.0f;
	}
	//==============s

	return shadow;
}

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation, float shadowFactor) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0f;
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return ((diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att) * (1.0f - shadowFactor);
}

void main() {
	vec3 n = normalize(normal_world);
	vec3 v = normalize(camera_world - position_world);
	vec3 r = reflect(-v, n);

	vec3 texColor = texture(diffuseTexture, uv).rgb;
	color = vec4(texColor * materialCoefficients.x, 1); // ambient
	vec3 color2 = color.xyz;

	float shadowFactor = CalcDirectionalShadowFactor(dirL);

	// add directional light contribution
	color.rgb += phong(n, -dirL.direction, v, dirL.color * texColor, materialCoefficients.y, dirL.color, materialCoefficients.z, specularAlpha, false, vec3(0), shadowFactor);

	// add point light contribution
	color.rgb += phong(n, pointL.position - position_world, v, pointL.color * texColor, materialCoefficients.y, pointL.color, materialCoefficients.z, specularAlpha, true, pointL.attenuation, 0.0);
	float reflectionFactor = mix(1.0, 0.2, shadowFactor)*clamp(color2.z-color2.x, 0.0, 1.0);

	color.rgb += reflectionFactor*10.0 * textureLod(EnvMap, r, pow(1.0-reflectionFactor, 4.0)*4.0).xyz;

}
